# CSSMinMaxR24.py for Notepad++ Python Scripting plugin * For CSS developers and web designers
# This is improved Python port of the YUI CSS Compressor featured to minify/format modern CSS
#
## Extends 2024 (R24)
# Patched by Barmaley: bugs fixed + useful features added * GPLv3 * GoodCRX Project
#
# 2024.8.31 (initial fork):
# 1. No required space for @media ") or (" = fixed.
# 2. No required spaces in calc with "+" sign 'calc(1px + 1em)' = fixed.
# 3. Preserved comments now defined with /**..*/ (instead of original /*!..*/).
# 4. If some text selected in active file: process fragment minification in current document (as in original script, see Usage below).
# 5. If nothing selected in active file: force process minification in NEW document (create new than process the copy of source CSS).
# 6. After processing autosave* new compressed file, autonamed from original name with change to "_.css".
#	 * Note: if any file with same name exist in the same dir - it's overwrite without confirmation.
# 7. Force pre-save current focument backup copy* (create in same dir as opened file, add .BAK to current filename).
#	 * This copy overwrite every time. No way in NPP to automatically remove temp backup files. Delete them manually if need.
#
# 2024.8.30 -> 2024.8.31:
# + Compressed CSS: all preserved comments delimits by new line from previous/following CSS rules (\n).
# + Fix logic: reset UNDO ops ONLY if minified "by place" (if process run in current file, else undo-buffer stay invalid).
# + Fix/improved: regexp for "calc(x + y)" expressions and for and/or conditions (e.g. @media/@import etc.)
# + Improve/optimize: multi-to-one zeros conversion, multi-semicolon remover, zero-value with units (px/vh/..) fixer
#
# 2024.8.31 -> 2024.9.4:
# + Rebrand project: nows it's named as CSSMinMaxR24
# + Simplest CSS autoformatter: autodetect compressed CSS then expand it - every rules converted to indented CSS blocks:
# 	* If some text selected: replace formatted fragment "by place" (as compressed in previous versions)
# 	* If nothing selected: current compessed CSS expands, create new autonamed file with *_MAX.css as name
#	* Yet works fine only with usual and one-level conditional rules (e.g. @media), otherwise text indents may be not "beauty"
# + Fix/improved: regexp for "calc(x + y)" expressions and for and/or conditions (e.g. @media/@import etc.)
# + Fix/improved: check/autofix for missing/excess :: in pseudo-elements (::after/etc) + fix excess :: in pseudo-classes (:hover/etc)
# + Improve/optimize: multi-to-one zeros conversion, multi-semicolon remover, zero-value with units (px/vh/..) fixer.
# + Simplest "CSS-beautifier": add some spaces after/before delimiters (e.g. ,_ :_ _{ ), no trailing spaces and empty lines.
# + Source code optimization: some "one-time-call" function moved to aproriate execution points.
#
# 2024.9.4 -> 2024.9.7:
# + Comment empty rules in compressor: rules with no declarations safe "self-backuped", no remove em (it will not processed by browser).
# + Fix/improved: check/autofix for missing/excess :: in pseudo-elements and excess ::+ in pseudo-classes (:hover/etc).
# + Post-process usability: after compress/expand - scroll to top if new file created or re-select converted fragment in current file.
# + Wrap-lines in compressor: get output CSS structure u want ("brick"/rules-by-lines/etc, optional, unused in source cssmin).
# + Add options block: u can check/change the "set{...}" params to adjust actions (read comments to understand param values).
#	Most of all options adjust the CSS-compressor (to minimize output file size), CSS-expander not required adjustment in general.
#	With changes the option flags you can:
#	* on/off initial backup (force save copy of current document);
#	* control method of empty rules operations* (keep/comment/remove);
#	* set your own "important" comments* mark (if u want preserv some comments), skip comments in actions (as-is) or force remove em;
#		* note: empty rules and comments can be removed ONLY by compressor, expander (CSS-formatter) NOT remove any trash anyway;
#	* control apply experimental features (e.g. pseudo-elements/classes correction);
#	* on/off wrap lines mode for compressor and/or beautifier for expander post-processing (improve readability);
#	* set your own mask of compressed/expanded CSS-file (when process entire document and autosave results in a new file);
#	* set every time autosaving before processing, intend only for converts "by place" (if works with selected text).
#
# 2024.9.7 -> 2024.9.15:
# + Fix/foolproof: convert direction autodetector fail correct detection in some case (for unlogic option flags combination). Fixed.
# + Fix/improved: check/autofix for missing :: in pseudo-elements (now check for all current MDN pe-items).
# + Fix/improved: processing of and/or in at-rules and math signs in calc() - check/fix whitespaces before/after that (both compressor/expander).
# + Force remove empty comment like /* */, anyway (both compress/expand ops, not intends of comments flag u set).
# + Force* move a @charset or @import rules to the BOF (both compress/expand ops). All @charset rules except the first will be force removed!
#	* See at //developer.mozilla.org/docs/Web/CSS/@charset and //developer.mozilla.org/docs/Web/CSS/@import for details.
# + Improved: conditional rules (CR) formatter now works fine with most* of multi-level CR (indents for @media/@supports/etc. inner rules).
#	* Note: sometimes u can get "crooked fir". It's possible in multi-level 4+ CR. Not bring any bug in expanded CSS, but sign to check your rules.
# + Optional: on/off multi-level* format of expanded CSS (expander only). If on - set extra indents in confitional at-rules (@media/@supports/etc).
#	* If off - use only basic CSS-format, indents will set only for in-rules declares, any embedded conditional rules processed as usual rules.
#	* As result - u get simplest one-level output CSS tree. Most fast processing with no iteratable steps to indent the inner conditions.
# + Optional: comments indent - compressor only (expander always do that), if on - all comments in output CSS will be indented by new line.
# + Optional: first comment protection, keep first* comment (CSS "cap") both in compressor/expander, stay unchanged and not removed anyway.
#	* Placed at BOF and NOT empty! Otherwise comment NOT consider as "first". Any preceding empty lines, tabs, whitespaces be stripped.
# + Optional: extra-fine compressor feats (converts multi-zeros to one, RGB-colors to HEX, shorten HEX colors, etc.) also can be used* in expander.
#	* Get the less size of expanded CSS and correct some typical exceeds (e.g. multi whitespaces, ";", etc.). If use - u get fine CSS at one pass.
#	* Experimental option (set[psel]) now works in pair with this. Check/autofix pseudo-elements/classes now execute ONLY if "extra-fine" flag is ON.
# + The Box Model Hack processing in compressor now is optional* step (this can be used ONLY for old MSIE-5/6, useless for modern browsers).
#	* Default is off (set[bmhk]:0), two exseed cycles in compressor excluded, faster processing. Set on ONLY if your CSS has such fixes for IE5/6.
# 	* See more about pseudo-class fix here: //tantek.com/CSS/Examples/boxmodelhack.html (RU: //www.webmascon.com/topics/coding/38a.asp).
# + Add apropriate options flags to new optional features: u can on/off em in the "set{...}" params block.
# + Misc: some code optimization, simplify/shorten some functions, add exseed [.#>] corrector(rule definitions), add test CSS-file in package.
#
## Requirements
#	0. NPP 6.5.1+ (this script tested in, older/newer NPP version - test yourself).
#	1. NPP Python Script plugin: install FULL version from //npppythonscript.sourceforge.net (this script runtime, required).
#	2. NPP Customize Toolbar: install from //sourceforge.net/projects/npp-customize (best usability if u often use this, optional).
#
## Installation
#	Simply copy "CSSMinMaxR24.py" to "..\plugins\PythonScript\scripts\" in your NPP folder and restart NPP.
#	It will now appear as a menu item under Plugins > Python Script > Scripts
#	For quick access u can adjust NPP Python Script plugin and add this script at top level in menu and/or on NPP toolbar.
#	For more usability u can also set the icon for this script - e.g. CSSMinMaxR24.bmp (supplied in package).
#	For best usability u can also use* the Customize Toolbar plugin to adjust menu button position/tooltip.
#	* Add the string "Plugins,Python Script,Scripts,CSSMinMaxR24,CSSMinMaxR24.bmp" in CustomizeToolbar.btn file.
#
## Usage
#	1. Select text fragment to process, or select nothing. Hint: select all (Ctrl+A) to process the whole current file "by place".
#	2. Go to NPP menu: Plugins..Python Script..Scripts..CSSMinMaxR24 and click. If u set the script toolbar button - simply click it.
#	3. Case some text selected, it will process and replace the selected fragment, otherwise the entire contents of current document.
#	4. Built-in NPP undo is available anyway - if u don't like the results. Also see options below to control autosave/undo actions.
#
## Hints
#	1. First time run (if your CSS yet not processed by this script) - run this script TWICE (to pass both compress/expand ops for your CSS).
#		* This is "control shot" to be sure any (possible) errors corrected. Simply select ALL text and apply script twice to that.
#	2. Use conversion without "extra-fine" processing only if u sure* your CSS has no errors. So, the process will passed more faster.
#		* If not - use the additional extra-fine conversion, process is some slower, but typical "kids" errors will be corrected.
#	3. Before "stream processing" all of your CSS files: test/verify conversions to be sure the script output is correct for your CSS.
#		* If u don't like some results - simply try change some settings, adjust processing flags as u wish and then test again.
#	4. Also u can "train on cats" with test CSS-file (included in package) to learn script works and try different options.
#	5. For more quality warranty: u can check the output CSS for errors online (e.g. at W3C CSS validator).
#
## Warning!
#	Don't switch to another tab in NPP during process runtime! Always wait for process end! Otherwise output CSS placed in tab u switch!
#	This unpredictable change is very possible on slow PC and if u work with big CSS (long time processing).
#
## Credits
#	Original cssmin.py port of YUI from //github.com/zacharyvoase/cssmin
#	Forked CSSmin.py v.1.0 from //github.com/ethanpil/npp-cssmin
#	W3C CSS validator //jigsaw.w3.org/css-validator/
#	Regex checker/validator //regex101.com/
#
## Licence
#	Original BSD changed to GPLv3 for this script and it's next childs.
#
###

## Intro ended. Are you ready? Let's go! :)
# Set options: conversion parameters, if not explained - 1 = on, 0 = off.
# Default values tested in practics, don't change em randomly if u not understand the process logic!
set = {	# NB! Four flags below are interdepends! Empty rules (if preserved) marks as "important" comments! Be sure u not lost em on changes!
	"rmcm": 1,	# Skip, process or force remove any comments. 0: skip (stay as-is), 1: automatic with preserv (default), 2: kill em all!
	"nlcm": 0,	# Comments indent with new line before/after for best readability. In compressor only, expanded CSS always has indents.
	"ercm": 2,	# Save empty rules as comments. 3: automatic, 2: save all, 1: only with all commented declares, 0: remove all trash.
				# Note: if set 3 (auto) - any empty rules or commented declares will be processed by comment rules logic (selective remove).
	"imcm": "",	# Preserv important (for you) comments from remove in compressor. Default "*" (or any other symbol) = keep that.
				# DON'T USE the regexp control symbols! Set any of [^.(){}\[\]!$+/?|] break the comments processing! (and will be skipped).
	"bckp": 0,	# Create current CSS file backup before any actions.
	"pfcm": 1,	# Keep first comment: both compressor/expander, that comment (CSS "cap") always stay unchanged and not removed anyway.
	"mlvl": 1,	# Multi-level format (expander only). 1: use extra indents for embedded rules in conditional at-rules (@media/@supports/etc).
	"fine": 1,	# Experimental: CSS-beautifier, for expander only, increase readability, remove exseed trash (empty lines, unleased ";", etc.)
	"extr": 1,	# Experimental: compressor's extras in expander, less size, increase readability, no exseeds (RGB/HEX-colors, multi-zeros, etc.)
				# Note: this flag intends both compressor/expander, if 0: any extra feats NOT be applied anyway (some increased output CSS size).
	"psel": 1,	# Experimental: try* to check/correct missing/exceed "::" for pseudo-elements and exceed ":" for pseudo-classes.
				# NB! This flag is work ONLY when "extra-fine" flag is on (this experimental features is one of the "code-correctors").
	"bmhk": 0,	# Use the Box Model Hack in compressor. Set to 1 ONLY if your CSS has this fix (for old IE5/6, useless for modern browsers).
	"tsas": 0,	# Autosave file every time before run process. Only for "by place" ops (if works with selected text in current file).
	"undo": 0,	# Use new undo buffer (for processing "by place" only). Use 1: ONLY if u sure - avoid reset current document undo buffer!
	"wrap": 0,	# Wrap lines in compressed CSS to an approximate length (increase CSS readability). 0 = no wrap, any integer = max length.
				# Try different values to get comfortable text view - e.g. 1: very fine - one line = one rule, 0: 1-line "brick", and so on.
	"fmax": "",	# Default file extension prefix mask for autosaved EXPANDED files. If empty - autoset as "_MAX", end ".css" be autoadded.
	"fmin": ""	# Default file extension prefix mask for autosaved COMPRESSED files. An mass can be ".min", end ".css" will be autoadded.
}

if set["bckp"]: notepad.saveAsCopy((notepad.getCurrentFilename() + ".BAK"))	# At first, create source file backup...if u ON this flag... :)
tmp = [[""],"","","","","",[""]]	# Array of temporary strings (used/changed/erased in both minifier/expander)

from StringIO import StringIO # The pure-Python StringIO supports unicode.
import re	# Regex ops module

# Remove CSS comments (except preserved).
def remove_comments(css):	# Skip user-defined comments sign if "preserv" mask is incorrect.
	if re.search(r'[\^\.\(\)\{\}\[\]\*\!\$\+\/\?\|\\]', set["imcm"]): set["imcm"] = 0	# Yet set as default "*" - to avoid exseed checks in proc
	preserve = iemac = 0
	deficm = "*"	# Default mark of preserved comments, also used to preserve empty rules from remove in CSS-compressor.
	comment_start = css.find("/*")
	while comment_start >= 0:	# Slicing to avoid an IndexError.
		if set["rmcm"] < 2: # Preserve comments like /**...*/ (by default: two stars ** followed open slash).
			chkpoint = css[comment_start + 2:comment_start + 3]
			preserve = chkpoint == deficm
			if not preserve and set["imcm"]: preserve = chkpoint == set["imcm"]
		comment_end = css.find("*/", comment_start + 2)
		if comment_end < 0:
			if not preserve:
				css = css[:comment_start]
				break
		elif comment_end >= (comment_start + 2):
			if css[comment_end - 1] == "\\":	# Check for IE Mac specific comment, leave this one and the following one alone.
				comment_start = comment_end + 2
				iemac = 1
			elif iemac:
				comment_start = comment_end + 2
				iemac = 0
			elif not preserve: css = css[:comment_start] + css[comment_end + 2:]
			else: comment_start = comment_end + 2
		comment_start = css.find("/*", comment_start)
	tmp[1][1] = "___PRESCOMM___"	# Preserved comments dumb, replaced to \n at last
	if set["imcm"]:
		css = re.sub(r"(.+)?\/\*"+set["imcm"], tmp[1][1]+"/*"+set["imcm"], css)
	return re.sub(r"^"+tmp[1][1], "", re.sub(r"(.+)?\/\*\*", tmp[1][1]+"/**", re.sub(r"\*\/\s*", "*/"+tmp[1][1], css)))

# Remove excess whitespace characters.
def remove_excess_whitespace(css):
	def pseudoclasscolon(css):	# Prevents 'p :link' becoming 'p:link'
		tmp[1][2] = "___PSEUDOCLSCOLON___"	# Dumb for 'p :link'
		rex = re.compile(r"(^|\})(([^\{\:])+\:)+([^\{]*\{)")
		m = rex.search(css)
		while m:
			css = ''.join([
				css[:m.start()],
				m.group().replace(":", tmp[1][2]),
				css[m.end():]])
			m = rex.search(css)
		return css

	css = re.sub(r"\s+([!{};:>\(\)\],]|\+(?!(\s+[\(\d-])))", r"\1", pseudoclasscolon(css))	# Remove spaces before things
	css = re.sub(r"([!{}:;>\(\[,]|\)?[^\s]\+)\s+", r"\1", css)	# Remove spaces after things
	return css.replace(tmp[1][2], ':')	# Put the colons back

# Convert rgb(51,102,153) to #336699.
def normalize_rgb_colors_to_hex(css):
	rex = re.compile(r"rgb\s*\(\s*([0-9,\s]+)\s*\)")
	m = rex.search(css)
	while m:
		colors = map(lambda s: s.strip(), m.group(1).split(","))
		hexcol = '#%.2x%.2x%.2x' % tuple(map(int, colors))
		css = css.replace(m.group(), hexcol)
		m = rex.search(css)
	return css

# Replace :0 0 0 0;, :0 0 0; etc. with :0;.
def condense_multidimensional_zeros(css):
	css = re.sub(r"\b:(\s*)?(0\s+){1,3}0;", ":0;", css)
	css = css.replace("background-position:0;", "background-position:0 0;")	# Revert to valid value.
	return css

# Shorten colors from #AABBCC to #ABC where possible.
def condense_hex_colors(css):
	rex = re.compile(r"([^\"'=\s])(\s*)(#+)([0-9A-F])([0-9A-F])([0-9A-F])([0-9A-F])([0-9A-F])([0-9A-F])", re.I)
	m = rex.search(css)
	while m:
		one = (m.group(4) + m.group(6) + m.group(8)).lower()
		two = (m.group(5) + m.group(7) + m.group(9)).lower()
		if one == two:
			css = css.replace(m.group(), m.group(1) + m.group(2) + '#' + one)
			m = rex.search(css, m.end() - 3)
		else: m = rex.search(css, m.end())
	return css

# Wrap the lines of the given CSS to an approximate length.
def wrap_css_lines(css, line_length):
	lines = []
	line_start = 0
	for i, char in enumerate(css):
		if char == '}' and (i - line_start >= line_length):	# It's safe to break after } characters.
			lines.append(css[line_start:i + 1])
			line_start = i + 1
	if line_start < len(css): lines.append(css[line_start:])
	return '\n'.join(lines)

# Try automatic fix for missing/excess :: (for rule selectors only).
def pseudo_elements_sign(css):
	tmp[2][1] = "(after|before|marker|selection|placeholder|backdrop|highlight|cue|file-selector-button|highlight|part|slotted|target-text|first-(letter|line)|(grammar|spelling)-error|view-transition(-(image-pair|group|new|old))?)"
	tmp[2][2] = "(?=.*\{)"
	css = re.sub(r"\s?[^:]:("+tmp[2][1]+"[^\w-])"+tmp[2][2], r"::\1", css)	# Pseudo-elements
	css = re.sub(r"(:\s*)+("+tmp[2][1]+"[\w-])"+tmp[2][2], r":\2", css)		# Pseudo-classes
	css = re.sub(r"(:\s*)+((?!"+tmp[2][1]+"))"+tmp[2][2], r":\2", css)		# Pseudo-classes
	css = re.sub(r"\s?:{2,}\s*", "::", css)	# Pseudo-elements above can not be ripped from parent
	return re.sub(r":+((-webkit-|-moz-)[\w-])"+tmp[2][2], r"::\1", css)	# Also check non-standard pseudo-elements

# Advanced CSS-processing: convert some declares to it's functional equivalent. Also correct some typical errors.
def extra_fine(css):	# NB! Apply ONLY if extras flag is ON (set[extr]=1)!
	tmp[2][0] = ["(%|px|vh|vw|vmin|vmax|em|rem|cm|mm|in|pt|pc|ex|ch|)", "([\+\*\/-])", "(?=[\(\d])"]
	if re.search(r"\b(and|or|calc)\s*\(", css):	# Add whitespaces in at-rules/calc() if omit
		css = re.sub(r"(\b(and|or)|"+tmp[2][0][1]+")\(", r" \1 (", css)	# In at-rules/calc()
		css = re.sub(r"\b(calc\s*\(.+(\d+"+tmp[2][0][0]+"|\)*))"+tmp[2][0][1], r"\1 \4 ", css)	# In calc(.)+(.): //developer.mozilla.org/docs/Web/CSS/calc
		css = re.sub(r"([\(-]*\d+"+tmp[2][0][0]+")"+tmp[2][0][1]+tmp[2][0][2], r"\1 \3 ", css)	# In calc((.+.))
		if not tmp[1][0]:	# Compressor only
			css = re.sub(r"(?=calc\s*\(.+(\d+"+tmp[2][0][0]+")|\)*)\s+([\*\/])\s+", r"\3", css)	# Minify for *|/
	css = re.sub(r"([\s:])(0)"+tmp[2][0][0], r"\1\2", css)	# Replace 0(px/em/%/etc) with 0.
	css = condense_multidimensional_zeros(css)
	css = re.sub(r"(:|\s)0+\.+(\d+)", r"\1.\2", css)	# Replace 0.5 to .5 if possible
	css = normalize_rgb_colors_to_hex(css)
	if set["psel"]: css = pseudo_elements_sign(css)	# Fix pseudo-elements rules
	if tmp[1][0]: css = re.sub(r'(([\t\s(]+[\w-]+:)(?=[^{]+[;)]?)|,)', r'\1 ', css)	# Expander only: add whitespace after ":"/"," in declares
	return condense_hex_colors(css)

# CSS compressor, remove indents from expanded CSS, minimize file size.
def cssmin(css):
	tmp[1] = [0,"","",""]	# Preset compressor flags/data
	if set["rmcm"]: css = remove_comments(css)	# Remove CSS comments (except preserved).
	if set["bmhk"]:	# Process the Box Model Hack (BMH): only for old MSIE-5/6 and if the hack presents in CSS
		tmp[1][3] = ['"\\"}\\""', "___PSEUDOCLASSBMH___"]	# BMH mask and pseudo-class dumb
		css = css.replace(tmp[1][3][0], tmp[1][3][1])
	css = remove_excess_whitespace(css)
	if set["rmcm"]: css = re.sub(tmp[1][1], "\n"*set["nlcm"], css)	# Return back the preserved comments with new line indent.
	if set["bmhk"]: css = css.replace(tmp[1][3][1], tmp[1][3][0])	# Revert back the BMH mask
	return css

# Basic CSS formatter, parse/indents as "usual" rules only.
def base_indent(css):
	tmp[1] = [1, "___AT_COND___", "___AT_RULES_SET___", "___AT_RULES_END___", r"([\"']data:)(.+;)(base64,)"]	# Expander flags/data
	css = re.sub(r'(\*\/)', r'\1\n', css)	# Comments with new lines
	if set["mlvl"]:	# Set dumbs for multilevel conditional at-rules (CR)
		css = re.sub('@', tmp[1][1], css)	# CR declare
		css = re.sub(r'('+tmp[1][1]+'[^\{]+)\{', r'\1'+tmp[1][2]+'\n;\n', css)	# CR begin
		css = re.sub('}}', ';\n'+tmp[1][3]+'\n', css)	# CR end
	else:
		css = re.sub(r'(@[^\{]+\{)', r'\1\n', css)	# Basic only: CR begin
	css = re.sub(r'([^({})]+\{)', r'\1\n\t', css)	# Rules begin
	if set["fine"] and re.search(tmp[1][4], css): css = css_beauty(css, 1)	# Add space before/after "!{,+>" and tabs base64-urls
	css = re.sub(r'(;((\s*\*\/)|(\s*\*\/\s*\w)))', r'\1\n\t', css)	# Next declare (also in-rule comments)
	css = re.sub(r'(?!"\'[^;]*);(?![^;]*\'")', r';\n\t', css)		# Next declare
	if set["fine"]: css = css_beauty(css, 2)	# Add spaces, indents, absent ";", process in-rule comments (best readability)
	css = re.sub(r'(\t*\})', r'\n}\n', css)	# Rules end
	return css

# Multi-level conditional rules (CR) formatter (@media/etc. indents for inner rules).
def cond_indent(css):
	def rule_shift(css, arr, c1, c2):
		shift = rshift = '\t'
		ishift = ""
		for rule in arr:	# Process rules inside the CR
			bc = re.search(c1, rule)
			ec = len(re.findall(c2, rule))
			if not bc:	# Skip "usual" rules with no CR.
				continue
			elif bc and not ec:	# Add tab-shift to next level inner rules
				css = css.replace(rule, rule.replace('\n', '\n'+shift)+'\n'+shift)
				ishift = shift
				shift += rshift
			elif ec:	# Build tab-shifted "fir" from rules inside CR
				tr = rule
				m = ec - 1
				for sr in re.split(c2, rule):
					if re.search(c1, sr):
						tr = tr.replace(sr, sr.replace('\n', '\n'+shift))+'\n\n'
					elif ec == m:
						tr = tr.replace(sr, sr.replace('\n', '\n'+ishift))+'\n'
					else:
						tr = tr.replace(sr, sr.replace('\n', '\n'+rshift*ec))+'\n'
					ec -= 1
					if not ec: break	# Omit shifts after CR end
				css = css.replace(rule, tr)
				shift = rshift
				ishift = ""
		return css

	css = rule_shift(css, re.split(tmp[1][1], css), tmp[1][2], tmp[1][3])
	css = re.sub(tmp[1][2], ' {', re.sub(tmp[1][1], '@', css))	# Revert back CR declares/begins/ends
	css = re.sub(r'(\n\s*)'+tmp[1][3], r'\n\t}\n}\n', re.sub(r'(\n\t*)\t?\t'+tmp[1][3], r'\1\t}\1}\n', css))
	return css

# CSS beautifier: increase output CSS-text readability, remove exceeds (both compressor/expander).
def css_beauty(css, step):
	if step == 1:	# Pre-improve when css-format is basic only (avoid unwanted issues at next process steps)
		css = re.sub(tmp[1][4], r'\1 \2\n\t\t\3', css)	# Experimental: add space after :/, in base64-urls
	if step == 2:	# Pre-improve before run conditional formatter, after all basic formats applied (avoid issues at next steps)
		css = re.sub(r"([>+])", r" \1 ", css)	# Add spaces before/after things.
		if not set["extr"]: css = re.sub(',', ', ', css)	# Add spaces after ","
		css = re.sub(r'([^\};]*)(?!\*\/)[\n\s]*\}', r'\1;\n}', css)	# Add absent ";" at last declaration end
		css = re.sub(r'(\{|!)', r' \1', css)	# Add spaces before {/!
		css = re.sub(r'(\n.+;)\s*\*\/', r'\1 */', css)	# Commented declares in rules
	else:	# Final readability improvments before autosave new file or replace selected fragment in current
		css = re.sub(r'[ ]{2,}', ' ', re.sub(r'([\.#]){2,}(?=.+[\{])', r'\1', css))	# Remove exceed whitespaces/dots/#
		if tmp[1][0]:
			css = re.sub(r'\n([^;\t]+;)(\s*\})', r'\n\t\1\2', css)	# Next declare after comments inside the rules
			css = re.sub(r'(\n\s*;*)+\n', r'\n', css.strip())	# Remove exceed ";", empty lines, trailing spaces/tabs
		else:
			css = re.sub(r";+\}", "}", re.sub(r";+", ";", css))	# Condense multiple ";;;" to one, remove trails
			if set["wrap"]: css = wrap_css_lines(css, set["wrap"])
			else: css = re.sub(r"\s+", " ", css)	# Condense multiple whitespaces/tabs to one
	return css.strip()+"\n"	# Remove trailing spaces, add line at the EOF

# CSS expander/formatter: expand compressed CSS for best readability.
def cssmax(css):
	css = base_indent(css)	# Basic formatter, one-level "usual" rules
	if set["mlvl"] and re.search(tmp[1][1], css): css = cond_indent(css)	# Check for and run CR-formatter
	return css

# Remove or preserv empty rules (safe "self-backup", no remove).
def empty_rules(css):
	r1 = r2 = ""
	if set["ercm"] > 1: r1 = r"\1/*here-was:empty-rule;*/}"
	if set["ercm"] > 0: r2 = r"/**\1\2\3*/"
	css = re.sub(r"([^}{]+{)([\s;]*)(})", r1, css)	# Empty rules with no declares (has selector only)
	css = re.sub(r"([^}{]+{[\s]*)[/*]+([^/*]*)[*/]+([\s]*})", r2, css)	# Rules with all commented declares
	return css

# Move special at-rules to BOF.
def move_to_begin(css, step):
	if step:	# If any special rule exist: temp store em and remove from source CSS before processing (will be moved to the BOF at the end).
		tmp[0] = ["", "___AT_CHARSET___", [], "___AT_IMPORT___", []]
		css = re.sub('@charset', tmp[0][1], css)
		css = re.sub('@import', tmp[0][3], css)
		atc = re.findall(r""+tmp[0][1]+" \"[^\"]*\";*", css)
		if len(atc):
			tmp[0][2] = atc[0]	# If there is any @charset rule, it's ONLY allow ONE and must be placed at the BOF.
			for ac in atc: css = css.replace(ac, "")	# Temp remove all @charset, only first will be revert at last.
		ati = re.findall(r""+tmp[0][3]+"\surl.+;*", css)
		if len(ati):
			tmp[0][4] = ati
			for ai in ati: css = css.replace(ai, "")	# Temp remove all @import, will be revert at last.
	else:	# Revert back special rules and preserved first comment (if any exist), force place whole block to the BOF.
		atc = ati = ""
		if len(tmp[0][2]):
			atc = re.sub(r"(\n*\s*;)+", ";", (re.sub(tmp[0][1], "@charset", tmp[0][2])+";\n"))
		if len(tmp[0][4]):
			lbr = tmp[1][0] or int(bool(set["wrap"]))
			for ai in tmp[0][4]: ati += ai+";"
			ati = re.sub(r";+", r";"+"\n"*lbr, re.sub(tmp[0][3], "@import", re.sub(r"(\n*\s*;)+", ";", ati)))+"/***/\n"
		css = atc + css + ati
	return css

# Preprocess by flag max|min.
def outcss(css, asf, cnf):
	def css_cap_back(css, cap):	# Revert back first comment and move a @charset/@import rules to the BOF (if any exist) after process ended.
		if cap: cap = cap[0].strip()+"\n"	# Revert back first comment (if exist).
		else: cap = ""
		if len(tmp[0])>1: cap = move_to_begin(cap, 0)	# Move a @charset/@import rules to the BOF (if any exist).
		return cap + css

	if set["extr"]: tmp[2] = ["","",""]	# Preset both process constants array for extra-fine conversions and pseudo-elements autofixer
	if asf: notepad.save()	# Autosave current file before processing (if set for selected text or force autosave if process all text in a new file).
	css = re.sub(r"\/\*[\s\*]*\*\/", "", css)	# Remove empty comments, anyway.
	if re.search("@(charset|import)", css): css = move_to_begin(css, 1)	# Search-and-store a @charset or @import rules.
	if set["pfcm"]:		# Keep first comment (CSS "cap") - always stay unchanged and not removed anyway.
		isfc = re.findall(r"^\s*\/\*[^/]+\*\/", css)	# Check for first comment exist.
		if len(isfc): css = css.replace(isfc[0], "")	# Temp store/remove original first comment to avoid any changes (placed back at the end).
	if set["ercm"] < 3:	css = empty_rules(css)	# Process empty rules
	# Autodetect direction (max|min) then run conversion (re.sub(...) = crutch to avoid incorrect run compressor on minified CSS with comments).
	max = re.search(r"\{\s*\n[^\/]", css) or not set["wrap"] and re.search(r"\}\n[^\/]", re.sub(r"(\{|\})\/", r"\1\n/", css))
	if max:	# Current CSS is expanded
		if cnf and set["fmin"]: tmp[0][0] = set["fmin"]	# Set output file prefix for CSS-minifier
		css = cssmin(css)
	else:	# Current CSS is minified
		if cnf:	# Set output file prefix for CSS-expander
			if set["fmax"]: tmp[0][0] = set["fmax"]
			else: tmp[0][0] = "MAX"
		css = cssmax(css)
	if set["extr"]: css = extra_fine(css)	# Advanced CSS-processing: convert some declares to it's short equivalent, correct some typical errors.
	elif re.search(r"\b(and|or|calc)\s*\(", css): css = re.sub(r"(\b(and|or)|\+|(\s-))(?=[\(\d])", r"\1 ", css)	# Force add whitespaces in at-rules/calc()
	css = css_beauty(css,0)	# General improve readability (final ops for both compressor/expander)
	if isfc or len(tmp[0])>1: css = css_cap_back(css, isfc)	# Revert back first comment and/or move a @charset or @import rules to the BOF.
	return css

## Functions defined. Now execute.
csstxt = editor.getSelText()	# Is selected text - process that, otherwise - process entire document
if len(csstxt):	# If selection in current document: process fragment "by place"
	tmp[6] = ["", editor.getSelectionStart(), ""]	# Usability: mem the start selection before conversion
	if set["undo"]:	editor.beginUndoAction()	# New undo buffer: reset current document undo buffer!
	editor.replaceSel(outcss(csstxt, set["tsas"], 0))	# Autosave current file before run process for selected text, every time (if set).
	if set["undo"]: editor.endUndoAction()
	# Usability: re-select converted fragment after conversion
	tmp[6][2] = editor.getCurrentPos()
	editor.setSelectionStart(tmp[6][1])
	editor.setSelectionEnd(tmp[6][2])
else:	# If nothing selected: set the new filename to create copy as new document
	tmp[6] = ["", notepad.getCurrentFilename(), outcss(editor.getText(), 1, 1)]	# Before process CSS in a new file - force autosave current file, anyway.
	notepad.new()
	editor.setText(tmp[6][2])
	notepad.saveAs(re.sub(r"\.css$", "_"+tmp[0][0]+".css", tmp[6][1]))
	editor.gotoLine(0)	# Usability: scroll to top after conversion

tmp = ""	# Reset temp-strings array

# That's all, folks! :)